import logging
from abc import abstractmethod
from dataclasses import dataclass
from functools import partial
from json.decoder import JSONDecodeError
from typing import Union

import requests

from pydispix.client import Client
from pydispix.color import Color
from pydispix.utils import resolve_url_endpoint

logger = logging.getLogger("pydispix")


@dataclass
class ChurchTask:
    x: int
    y: int
    color: Union[int, str, tuple[int, int, int], Color]


class ChurchClient(Client):
    def __init__(
        self,
        pixel_api_token: str,
        church_token: str,
        base_church_url: str,
        *args,
        **kwargs
    ):
        super().__init__(pixel_api_token, *args, **kwargs)

        if not base_church_url.endswith("/"):
            base_church_url = base_church_url + "/"

        self.base_church_url = base_church_url
        self.church_token = church_token

    def resolve_church_endpoint(self, endpoint: str):
        return resolve_url_endpoint(self.base_church_url, endpoint)

    @abstractmethod
    def get_task(self, endpoint: str = "get_task") -> ChurchTask:
        """
        Get task from the church, this is an abstract method, you'll need
        to override this to get it to work with your church's specific API.
        """

    @abstractmethod
    def submit_task(self, church_task: ChurchTask, endpoint: str = "submit_task") -> bool:
        """
        Submit a task to the church, this is an abstract method, you'll need
        to override this to get it to work with your church's specific API.
        """

    def _put_and_submit(self, church_task: ChurchTask, endpoint: str = "submit_task", show_progress: bool = False):
        """
        Put new pixel on the canvas and send the `submit_task` request.
        These are done together here because `put_pixel` waits for API
        rate limits from Pixels API, while the church rate-limit on
        task submitting might be lower than the api cooldown time.
        Because of this, the submit task is ran as `task_after` to
        `put_pixel` function.
        """
        submit_task = partial(self.submit_task, church_task, endpoint)
        self.put_pixel(
            church_task.x, church_task.y, church_task.color,
            show_progress=show_progress,
            task_after=submit_task
        )

    def run_task(self, show_progress: bool = False) -> bool:
        """Obtain, run and submit a single task to the church."""
        task = self.get_task()
        logger.info(f"Running church task: {task}")

        # Handle 403/400 by resetting the pixel, these happen when somebody
        # managed to overwrite the pixel we set before we informed the server
        # that we actually set it, it's quite rare, but it can happen.
        while True:
            try:
                return self._put_and_submit(task, show_progress=show_progress)
            except requests.HTTPError as exc:
                resp = exc.response
                if resp.status_code == 403 or resp.status_code == 400:
                    try:
                        log_detail = resp.json()
                    except JSONDecodeError:
                        raise exc
                    logger.warning(f"Repeating task, got 403: {log_detail}")
                raise exc

    def run_tasks(self, show_progress: bool = False):
        """Keep running church tasks forever."""
        while True:
            self.run_task(show_progress=show_progress)



from typing import Any
import requests
import re

from pydispix.ratelimits import RateLimitedEndpoint


class PyDisPixError(Exception):
    """Parent class for all exceptions defined by this library"""


class RateLimitBreached(PyDisPixError):
    """Request failed due to rate limit breach."""
    def __init__(self, *args, response: requests.Response, **kwargs):
        super().__init__(*args, **kwargs)

        # Get time limits from headers with RateLimitedEndpoint
        temp_rate_limit = RateLimitedEndpoint()
        temp_rate_limit.update_from_headers(response.headers)

        self.requests_limit = temp_rate_limit.requests_limit
        self.reset_time = temp_rate_limit.reset_time
        self.remaining_requests = temp_rate_limit.remaining_requests
        self.cooldown_time = temp_rate_limit.cooldown_time

        # Store the expected wait and the original response which trigerred this exception
        self.expected_wait_time = temp_rate_limit.get_wait_time()
        self.response = response

    def __str__(self):
        s = super().__str__()
        s += f"\nresponse={self.response.content}"
        if self.expected_wait_time != 0:
            s += f"\nexpected_wait_time={self.expected_wait_time}"

        return s


class InvalidToken(PyDisPixError, requests.HTTPError):
    """Invalid token used."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CanvasFormatError(PyDisPixError):
    """Exception raised when the canvas is badly formatted."""


class InvalidColor(PyDisPixError):
    """Invalid color format"""

    def __init__(self, *args, color: Any, **kwargs):
        super().__init__(*args, **kwargs)
        self.color = color

    def __str__(self) -> str:
        s = super().__str__()
        return s + f" color={self.color}"


class OutOfBoundaries(PyDisPixError):
    """Status code 422 - tried to draw a pixel outside of the canvas"""


def handle_invalid_body(response: requests.HTTPError) -> PyDisPixError:
    """
    Invalid body can mean many things, this analyzed given response
    and returns appropriate exception for it.
    """
    if response.status_code != 422:
        raise ValueError("Invalid Body response must have 422 HTTP code.")

    detail = response.json()['detail']

    # Work with 1st entry only, we can't raise multiple errors anyway
    entry = detail[0]
    if entry["loc"][1] == "rgb":
        color = re.search(r"'(.+)' is not a valid color", entry["msg"]).groups()[0]
        return InvalidColor("Couldn't resolve color", color=color)
    if entry["loc"][1] in ("x", "y"):
        return OutOfBoundaries(entry["msg"])

    raise requests.HTTPError("Unrecognized 422 exception, please report this issue in the pydispix repository", response=response)



import logging
import random
from dataclasses import dataclass
from typing import Callable, Optional, Union

from pydispix.church import ChurchClient, ChurchTask

logger = logging.getLogger("pydispix")

SQLITE_CHURCH = "https://decorator-factory.su"
RICK_CHURCH = "https://pixel-tasks.scoder12.repl.co/api"


@dataclass
class RickChurchTask(ChurchTask):
    project_title: str
    start: float


@dataclass
class SQLiteChurchTask(ChurchTask):
    id: int
    issued_by: str


class RickChurchClient(ChurchClient):
    """Church Client designed to work specifically with rick church"""
    def __init__(
        self,
        pixel_api_token: str,
        church_token: str,
        base_church_url: str = RICK_CHURCH,
        *args,
        **kwargs
    ):
        super().__init__(pixel_api_token, church_token, base_church_url, *args, **kwargs)

    def make_request(
        self, method: str, url: str, *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        parse_json: bool = True,
        ratelimit_after: bool = False,
        task_after: Optional[Callable] = None,
        show_progress: bool = False,
        repeat_on_ratelimit: bool = True,
    ) -> Union[bytes, dict]:
        """Extend regular make_request and add `key` parameter to the request"""
        if params is None:
            params = {}
        params["key"] = self.church_token

        return super().make_request(
            method, url,
            data=data,
            params=params,
            parse_json=parse_json,
            ratelimit_after=ratelimit_after,
            task_after=task_after,
            show_progress=show_progress
        )

    def get_task(self, endpoint: str = "get_task") -> RickChurchTask:
        url = self.resolve_church_endpoint(endpoint)
        response = self.make_request("GET", url)
        return RickChurchTask(**response["task"])

    def submit_task(self, church_task: RickChurchTask, endpoint: str = "submit_task") -> dict:
        url = self.resolve_church_endpoint(endpoint)
        body = {
            'project_title': church_task.project_title,
            'start': church_task.start,
            'x': church_task.x,
            'y': church_task.y,
            'color': church_task.color
        }
        return self.make_request("POST", url, data=body)


class SQLiteChurchClient(ChurchClient):
    """Church Client designed to work specifically with rick church"""
    def __init__(
            self,
            pixel_api_token: str,
            base_church_url: str = SQLITE_CHURCH,
            *args,
            **kwargs
    ):
        # SQLite Church API is open for everyone, it doesn't need a token
        church_token = None
        super().__init__(pixel_api_token, church_token, base_church_url, *args, **kwargs)

    def get_task(self, endpoint: str = "tasks") -> SQLiteChurchTask:
        url = self.resolve_church_endpoint(endpoint)
        response = self.make_request("GET", url)
        task = random.choice(response)
        return SQLiteChurchTask(**task)

    def submit_task(self, church_task: SQLiteChurchTask, endpoint: str = "submit_task"):
        url = self.resolve_church_endpoint(endpoint)
        body = {"task_id": church_task.id}
        return self.make_request("POST", url, data=body)



def resolve_url_endpoint(base_url: str, endpoint_url: str) -> str:
    """Add `endpoint_url` to the `base_url` and return the whole URL"""
    endpoint_url = endpoint_url.removeprefix("/")
    if endpoint_url.startswith("https://") or endpoint_url.startswith("http://"):
        if endpoint_url.startswith(base_url):
            return endpoint_url
        raise ValueError("`endpoint` referrs to unknown full url, that doesn't belong to the base url")
    else:
        return base_url + endpoint_url



import os
import requests
import logging
from collections import namedtuple
from typing import Union, Optional, Callable
from json import JSONDecodeError

from pydispix.ratelimits import RateLimiter
from pydispix.canvas import Canvas, Pixel
from pydispix.color import Color, parse_color
from pydispix.errors import RateLimitBreached, InvalidToken, handle_invalid_body
from pydispix.utils import resolve_url_endpoint

logger = logging.getLogger('pydispix')
Dimensions = namedtuple('Dimensions', ('width', 'height'))


class Client:
    """HTTP client to the pixel API."""
    def __init__(self, token: Optional[str] = None, base_url: str = "https://pixels.pythondiscord.com/"):
        if token is None:
            try:
                token = os.environ['TOKEN']
            except KeyError:
                raise RuntimeError("Unable to load token, 'TOKEN' environmental variable not found.")

        if not base_url.endswith("/"):
            base_url = base_url + "/"

        self.token = token
        self.base_url = base_url
        self.headers = {
            "Authorization": "Bearer " + token,
            "User-Agent": "ItsDrike pydispix",
        }
        self.rate_limiter = RateLimiter()

    def make_raw_request(
        self, method: str, url: str, *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        update_rate_limits: bool = True,
    ) -> requests.Response:
        """
        This method is here purely to make an HTTP request and update the rate limiter.
        Even though this will update the rate limtis, it will not wait for them.
        """
        logger.debug(f"Request: {method} on {url} {data=} {params=}.")
        response = requests.request(
            method, url,
            json=data,
            params=params,
            headers=self.headers
        )

        if update_rate_limits:
            self.rate_limiter.update_from_headers(url, response.headers)

        if response.status_code == 429:
            logger.debug(f"Request: {method} on {url} {data=} {params=} has failed due to rate limitation.")
            raise RateLimitBreached(
                "Request didn't succeed because it was made during a rate-limit phase.",
                response=response
            )
        if response.status_code == 401:
            logger.error("Request failed with 401 (Forbidden) code. This means your API token is most likely invalid.")
            raise InvalidToken("Received 401 - FORBIDDEN: Is your API token correct?", response=response)

        if response.status_code == 422:
            exc = handle_invalid_body(response)
            if exc is not None:
                raise exc

        if response.status_code != 200:
            raise requests.HTTPError(f"Received code {response.status_code}", response=response)

        return response

    def make_request(
        self, method: str, url: str, *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        parse_json: bool = True,
        ratelimit_after: bool = False,
        task_after: Optional[Callable] = None,
        repeat_on_ratelimit: bool = True,
        show_progress: bool = False,
    ) -> Union[bytes, dict]:
        """
        Make a `method` request to given `url`, while respecting the API rate limits.
        You can optionally pass JSON `data` and `parameters`.
        You can set `show_progress` to draw a progress bar while waiting for the rate limtis.

        For some requests, you may want to wait for the rate limit timeout *after* the request
        was made. This is needed because with some requests we want the most recent data we
        can get, after waiting out the limit (for example with get_canvas), but with others
        where we want to make our request as soon as possible, and only then wait for rate limits.

        You can also set `task_after` which is a callable function, that will be executed right
        after the request is made done. (This is needed if a user needs to report that he finished
        a set_pixel task to a church before waiting out the time limit). If this is used, you'll
        probably also want to have `ratelimit_after` enabled, to wait out the rate limit after
        the task with this `task_after` is sent.

        If `repeat_on_ratelimit` is set, the task will be re-run 1 more time. This is the default
        setting, since we can encounter initial rate limtis from the pixel API, and after we update
        the rate_limiter and wait out the proper cooldown, we make the request again. If the 2nd
        request fails too however, `RateLimitBreached` will be raised, to avoid infinite loops.
        """
        if not ratelimit_after:
            self.rate_limiter.wait(url, show_progress=show_progress)
        try:
            response = self.make_raw_request(method, url, data=data, params=params)
        except RateLimitBreached as exc:
            try:
                response_text = exc.response.json()
            except JSONDecodeError:
                response_text = exc.response.content
            # This can happen with first request when we're rate-limiting afterwards
            # Or when 2 machines are using the same token. When this occurs we continue
            # and make the request again.
            logger.warning(f"Hit rate limit, repeating request ({response_text})")

            # Wait for the rate limit if we're limiting afterwards
            # we don't send the `task_after` here, because the request failed anyway
            if ratelimit_after:
                self.rate_limiter.wait(url, show_progress=show_progress)

            if repeat_on_ratelimit:
                return self.make_request(
                    method, url,
                    data=data, params=params,
                    parse_json=parse_json,
                    ratelimit_after=ratelimit_after,
                    task_after=task_after,
                    show_progress=show_progress,
                    repeat_on_ratelimit=False
                )
            else:
                raise exc

        # Run the task as soon as we can, before waiting for any rate limitations
        if task_after:
            task_after()

        if ratelimit_after:
            self.rate_limiter.wait(url, show_progress=show_progress)

        if parse_json:
            return response.json()
        else:
            return response.content

    def resolve_endpoint(self, endpoint: str) -> str:
        """Resolve given `endpoint` to use the base_url"""
        return resolve_url_endpoint(self.base_url, endpoint)

    def get_dimensions(self) -> Dimensions:
        """Make a request to obtain the canvas dimensions"""
        url = self.resolve_endpoint("get_size")
        data = self.make_request("GET", url)
        return Dimensions(width=data["width"], height=data["height"])

    def get_canvas(self, show_progress: bool = False) -> Canvas:
        """Fetch the whole canvas and return it in a `Canvas` object."""
        url = self.resolve_endpoint("get_pixels")
        data = self.make_request("GET", url, parse_json=False, show_progress=show_progress)
        size = self.get_dimensions()
        return Canvas(size, data)

    def get_pixel(self, x: int, y: int, show_progress: bool = False) -> Pixel:
        """Fetch rgb data about a specific pixel"""
        url = self.resolve_endpoint("get_pixel")
        data = self.make_request("GET", url, params={"x": x, "y": y}, show_progress=show_progress)
        hex_color = data["rgb"]
        return Pixel.from_hex(hex_color)

    def put_pixel(
        self,
        x: int, y: int,
        color: Union[int, str, tuple[int, int, int], Color],
        show_progress: bool = False,
        task_after: Optional[Callable] = None
    ) -> str:
        """
        Draw a pixel and return a message.

        If you are reporting to some church, you may want to set
        `task_after` to the completion request for your church.
        This way, the task will be executed instantly, before
        waiting for the rate limitation
        """
        url = self.resolve_endpoint("set_pixel")
        data = self.make_request(
            'POST', url,
            data={
                'x': x,
                'y': y,
                'rgb': parse_color(color)
            },
            show_progress=show_progress,
            ratelimit_after=True,
            task_after=task_after
        )

        logger.info('Success: {message}'.format(**data))
        return data['message']
